#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
#
# This script is licensed under GNU GPL version 2.0 or above
# (c) 2024 Antonio J. Delgado
"""IMAP filter tool"""

import sys
import os
import stat
import logging
from logging.handlers import SysLogHandler
import imaplib
import email
from pathlib import Path
from signal import signal, SIGINT
import json
import re
import codecs
import time
from yaml import load
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
import click
import click_config_file

class ImapFilter:
    '''IMAP filter tool'''

    def __init__(self, **kwargs):
        start_time = time.time()
        self.config = kwargs
        if 'log_file' not in kwargs or kwargs['log_file'] is None:
            self.config['log_file'] = os.path.join(
                os.environ.get(
                    'HOME',
                    os.environ.get(
                        'USERPROFILE',
                        os.getcwd()
                    )
                ),
                'log',
                'imap_filter.log'
            )
        self._init_log()
        signal(SIGINT, self._signal_handler)
        with open(self.config['filters_file'], 'r', encoding='UTF-8') as filters_file:
            if '.json' in self.config['filters_file'].lower():
                self.config['mailboxes'] = json.load(filters_file)
            elif '.yaml' in self.config['filters_file'].lower() or '.yml' in self.config['filters_file'].lower():
                self.config['mailboxes'] = load(filters_file, Loader=Loader)
        if len(self.config['mailboxes']) == 0:
            self._log.error(
                "Filters file is empty. Use --help to see more details."
            )
            sys.exit(1)
        self.connect_imap()
        self._process_filters()
        self._log.debug(
            "Took %s seconds to process",
            time.time() - start_time
        )

    def _process_filters(self):
        matches = 0
        for mailbox in self.config['mailboxes']:
            if 'mailbox' not in mailbox:
                self._log.warning(
                    "No mailbox name in %s, skipping.",
                    mailbox
                )
                break
            if mailbox['mailbox'] == 'INBOX' and self.config['sieve_scripts_path']:
                self._create_sieve_script(mailbox['filters'])
            self._log.debug(
                "Processing mailbox '%s'...",
                mailbox['mailbox']
            )
            self.imap.select(mailbox=mailbox['mailbox'], readonly=False)
            self._log.debug("Searching for all messages in mailbox '%s'...", mailbox['mailbox'])
            typ, data = self.imap.search('UTF-8', 'UNDELETED')
            if typ != 'OK':
                self._log.error('Error, server replied: %s', data)
                return False
            all_msgs_uids = data[0].split()
            total_msgs = len(all_msgs_uids)
            self._log.debug(
                "Processing %s messages in mailbox '%s'...",
                total_msgs,
                mailbox['mailbox']
            )
            msg_count = 0
            self.mailbox_matches = 0
            for message_id in all_msgs_uids:
                msg_count += 1
                self._log.debug(
                    "Fetching message %s of %s...",
                    msg_count,
                    total_msgs
                )
                try:
                    typ, data = self.imap.fetch(message_id, '(RFC822)')
                    if typ != 'OK':
                        self._log.error('Error, server replied: %s', data)
                        return False
                    typ, unseen_data = self.imap.store(message_id, '-FLAGS', '\\Seen')
                    if typ != 'OK':
                        self._log.error('Error, server replied: %s', unseen_data)
                        return False
                    for mfilter in mailbox['filters']:
                        self._log.debug("Processing filter '%s'...", mfilter['name'])
                        if self._process_message(message_id, data[0], mfilter)['match']:
                            break
                except imaplib.IMAP4.error as error:
                    self._log.error(
                        "Error fetching message. %s",
                        error
                    )
            if self.mailbox_matches > 0:
                self._log.info(
                    "A total of %s matches for this mailbox,",
                    self.mailbox_matches
                )
            else:
                self._log.info(
                    "No matches for this mailbox"
                )
            matches += self.mailbox_matches
        self._log.info(
            "A total of %s matches for all mailboxes.",
            matches
        )
        if matches > 0:
            try:
                self.imap.expunge()
            except imaplib.IMAP4.abort as error:
                self._log.error(
                    "Error expunging connection. %s",
                    error
                )
        return True

    def _create_sieve_script(self, filters):
        if os.path.exists(self.config['sieve_scripts_path']):
            if os.geteuid() == 0:
                file_path = Path(self.config['sieve_scripts_path'])
                owner = file_path.owner()
                group = file_path.group()
            with open(self.config['sieve_scripts_path'], 'r', encoding='UTF-8') as sieve_script:
                content = sieve_script.read()
        else:
            content = '# Managed by imap_filter. Some lines might get replaced if the match with rule name from imap_filter'
            content += 'require ["fileinto"];'
        for mfilter in filters:
            self._log.debug(
                'Adding filter %s to sieve script',
                mfilter['name']
            )
            if 'field' in mfilter:
                field = mfilter['field']
            else:
                field = 'From'
            if 'regexp' in mfilter:
                condition = f'allof (header :contains "{field}" "{mfilter['regexp']}")'
            else:
                condition = f'allof (header :contains "{field}" ["{'", "'.join(mfilter['words'])}"])'
            search = f"^.*  # imap_filter: rule name 'imap_filter_INBOX_{mfilter['name']}'$"
            replacement = f'if {condition} {{ fileinto "{mfilter['destination']}"; stop;}}  # imap_filter: rule name \'imap_filter_INBOX_{mfilter['name']}\''
            match = re.search(search, content, re.MULTILINE)
            if match:
                self._log.debug(
                    "Line '%s' found in current file",
                    search
                )
                new_content = re.sub(search, replacement, content)
                if content != new_content:
                    self._log.debug(
                        "It's different, so replacing it with '%s'",
                        replacement
                    )
                    content = f"{new_content}\n"
            else:
                self._log.debug(
                    "Line '%s' not found in current file, adding it",
                    search
                )
                content += f"{replacement}\n"
        with open(self.config['sieve_scripts_path'], 'w', encoding='UTF-8') as sieve_script:
            sieve_script.write(content)
        if os.geteuid() == 0:
            os.chown(self.config['sieve_scripts_path'], owner, group)
        os.chmod(
            self.config['sieve_scripts_path'],
            0o777
            )
        return True

    def _process_message(self, message_id, data, mfilter):
        '''Process a mail message'''
        if isinstance(data[1], int):
            self._log.warning(
                "Unexpected response fetching message: '%s'. Try again.",
                data
            )
            return {"match": False}
        try:
            part = data[1].decode('utf-8')
        except UnicodeDecodeError:
            part = str(data[1])
        message = email.message_from_string(part)
        decoded_field = email.header.decode_header(message.get(mfilter['field'], ""))
        self._log.debug(
            "Decoded field: %s",
            decoded_field
        )
        if isinstance(decoded_field[0][0], str):
            field_data = decoded_field[0][0]
        else:
            field_data = decoded_field[0][0].decode()
        if 'words' in mfilter:
            regexp = '^(?=.*' + '.*|.*'.join(mfilter['words']) + '.*)'
        else:
            if 'regexp' in mfilter:
                regexp = mfilter['regexp']
            else:
                self._log.error(
                    "The filter '%s' doesn't have a 'words' or 'regexp' value. %s",
                    mfilter['name'],
                    mfilter
                )
                return {"match": False}
        self._log.debug(
            "Testing match for '%s' in '%s'...",
            regexp,
            field_data
        )
        match = re.match(regexp, field_data)
        if match:
            self._log.info(
                "Field '%s' => '%s', matches filter '%s'",
                mfilter['field'],
                field_data,
                mfilter['name']
            )
            self.mailbox_matches += 1
            if self.config['dummy']:
                self._log.info('Doing nothing (dummy run)')
            else:
                self._do_filter(message_id, mfilter)
            return {"match": True}
        return {"match": False}

    def _do_filter(self, message_id, mfilter):
        if f"_action_{mfilter['action']}" in dir(self):
            function = getattr(self, f"_action_{mfilter['action']}")
            # try:
            result = function(message_id, mfilter)
            self._log.debug(
                "Result: %s",
                result
            )
            # except Exception as error:
            #     self._log.error(
            #         b"Error performing '%s' action with message %s. Filter: %s. Error: %s",
            #         mfilter['action'],
            #         message_id,
            #         mfilter,
            #         error
            #     )

    def _action_move(self, message_id, mfilter):
        self._log.info(
            "Moving message '%s' to '%s'...",
            message_id,
            mfilter['destination']
        )
        if self._action_copy(message_id, mfilter):
            self._action_delete(message_id, None)

    def _create_mailbox(self, mailbox):
        self._log.debug(
            "Creating mailbox '%s'...",
            mailbox
        )
        typ, data = self.imap.create(mailbox)
        if typ != 'OK':
            self._log.error(
                'Error creating mailbox %s, server replied: %s',
                mailbox,
                data
            )
            return False
        return True

    def _action_copy(self, message_id, mfilter):
        self._log.info(
            "Copying message '%s' to '%s'...",
            message_id,
            mfilter['destination']
        )
        mailbox_encoded = codecs.encode(
            mfilter['destination'],
            encoding="utf-7"
        ).replace(b"+", b"&")
        typ, data = self.imap.copy(message_id, mailbox_encoded)
        if typ != 'OK':
            if b'[TRYCREATE]' in data[0]:
                if self._create_mailbox(mailbox_encoded):
                    typ, data = self.imap.copy(message_id,mailbox_encoded)
                    if typ != 'OK':
                        self._log.error(
                            'Error copying message to %s, server replied: %s',
                            mfilter['destination'],
                            data
                        )
                        return False
            else:
                self._log.error(
                    'Error copying message to %s, server replied: %s',
                    mfilter['destination'],
                    data
                )
                return False
        return True

    def _action_delete(self, message_id, mfilter):
        self._log.info(
            "Deleting message '%s'...",
            message_id
        )
        typ, data = self.imap.store(message_id, '+FLAGS', '\\Deleted')
        if typ != 'OK':
            self._log.error(
                'Error marking message %s as seen, server replied: %s',
                message_id,
                data
            )
            return False
        return True

    def _action_mark_seen(self, message_id, mfilter):
        self._log.info(
            "Marking as seen message '%s'...",
            message_id
        )
        typ, seen_data = self.imap.store(message_id, '+FLAGS', '\\Seen')
        if typ != 'OK':
            self._log.error(
                'Error marking message %s as seen, server replied: %s',
                message_id,
                seen_data
            )
            return False
        return True

    def connect_imap(self):
        '''Create connection object to the IMAP server'''
        self._log.debug(
            'Connecting to server %s:%s...',
            self.config['imap_server'],
            self.config['imap_port']
        )
        if self.config['ssl']:
            try:
                self.imap = imaplib.IMAP4_SSL(self.config['imap_server'], self.config['imap_port'])
            except Exception as error:
                self._log.error(
                    "Error connecting securely to IMAP server '%s'. %s",
                    self.config['imap_server'],
                    error
                )
                sys.exit(1)
        else:
            try:
                self.imap = imaplib.IMAP4(self.config['imap_server'], self.config['imap_port'])
            except Exception as error:
                self._log.error(
                    "Error connecting to IMAP server '%s'. %s",
                    self.config['imap_server'],
                    error
                )
                sys.exit(2)
        try:
            self._log.debug('Authenticating as user %s...', self.config['imap_user'])
            self.imap.login(self.config['imap_user'], self.config['imap_password'])
        except Exception as error:
            self._log.error("Error while login as '%s'. %s'", self.config['imap_user'], error)
            self.imap.close()
            sys.exit(3)

    def _signal_handler(self, signal_received, frame):
        # Handle any cleanup here
        self._log.info(
            'SIGINT or CTRL-C (%s %s) detected. Exiting gracefully',
            signal_received,
            frame
        )
        self.imap.close()
        self.imap.logout()
        sys.exit(0)

    def _init_log(self):
        ''' Initialize log object '''
        self._log = logging.getLogger("imap_filter")
        self._log.setLevel(logging.DEBUG)

        sysloghandler = SysLogHandler()
        sysloghandler.setLevel(logging.DEBUG)
        self._log.addHandler(sysloghandler)

        streamhandler = logging.StreamHandler(sys.stdout)
        streamhandler.setLevel(
            logging.getLevelName(self.config.get("debug_level", 'INFO'))
        )
        self._log.addHandler(streamhandler)

        if 'log_file' in self.config:
            log_file = self.config['log_file']
        else:
            home_folder = os.environ.get(
                'HOME', os.environ.get('USERPROFILE', '')
            )
            log_folder = os.path.join(home_folder, "log")
            log_file = os.path.join(log_folder, "imap_filter.log")

        if not os.path.exists(os.path.dirname(log_file)):
            os.mkdir(os.path.dirname(log_file))

        filehandler = logging.handlers.RotatingFileHandler(
            log_file, maxBytes=102400000
        )
        # create formatter
        formatter = logging.Formatter(
            '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'
        )
        filehandler.setFormatter(formatter)
        filehandler.setLevel(logging.DEBUG)
        self._log.addHandler(filehandler)
        return True


@click.command()
@click.option(
    "--debug-level",
    "-d",
    default="INFO",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        case_sensitive=False,
    ),
    help='Set the debug level for the standard output.'
)
@click.option('--log-file', '-l', help="File to store all debug messages.")
@click.option(
    "--dummy","-n", is_flag=True,
    help="Don't do anything, just show what would be done."
)
@click.option('--imap-server', '-s', default='localhost', help='IMAP server')
@click.option(
    '--imap-port', '-p', default=993,
    type=click.IntRange(1, 65535), help='IMAP server port'
)
@click.option(
    '--imap-user', '-u', required=True,
    help='User name to use for the connection to the IMAP server'
)
@click.option(
    '--imap-password', '-P', required=True,
    help='Password to connect to the IMAP server. Warning! Use a configuration file to avoid revelaing your passwords.'
)
@click.option(
    '--ssl', '-S', default=True,
    help='Whether to use a secure connection or not.'
)
@click.option(
    '--filters-file', '-F', required=True,
    help='JSON or YAML file containing a list of dictionaries with the filter rules.'
)
@click.option(
    '--sieve-scripts-path',
    '-P',
    help='Path to create sieve scripts.'
)
@click_config_file.configuration_option()
def __main__(**kwargs):
    return ImapFilter(**kwargs)


if __name__ == "__main__":
    __main__()
